// Copyright 2006-2007 all rights reserved; see LICENSE file for BSD-style license

package edu.berkeley.sbp;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.util.*;
import edu.berkeley.sbp.Parser.Table.*;
import edu.berkeley.sbp.Sequence.Pos;
import edu.berkeley.sbp.Sequence.Pos;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;

/** a node in the GSS */
final class StateNode
    implements Invokable<Pos, ResultNode, Object>,
               IntegerMappable,
               GraphViz.ToGraphViz,
               Iterable<ResultNode> {

    /** which GSS.Phase this StateNode belongs to */
    public GSS.Phase phase() { return phase; }
    public Iterator<ResultNode> iterator() { return results.iterator(); }
    public Parser.Table.State state() { return state; }

    public int toInt() { return idx; }

    boolean destroyed = false;

    public void check() {
        if (destroyed) return;
        boolean dead = true;
        // - all nodes must have a predecessor
        // - non-doomed nodes must either:
        //      - be on the frontier or
        //      - have a non-doomed node closer to the frontier than themself
        if (phase.isFrontier()) dead = false;
        else for(ResultNode r : successors)
                 if (state.doomed) { if (r.usedByAnyNode()) { dead = false; break; } }
                 else              { if (r.usedByNonDoomedNode()) { dead = false; break; } }
        dead |= results.size()==0;
        if (!dead) return;
        destroyed = true;
        if (phase() != null && phase().hash != null)
            phase().hash.remove(state, predPhase);
        while(successors.size()>0)
            for(ResultNode r : successors) {
                successors.remove(r);
                r.removePred(this);
                break;
            }
        while(results.size()>0)
            for(ResultNode r : results) {
                results.remove(r);
                r.removeSucc(this);
                break;
            }
        results = null;
        successors = null;
    }

    //////////////////////////////////////////////////////////////////////

    private static int node_idx = 0;
    private final int idx = node_idx++;

    private final GSS.Phase phase;
    private final GSS.Phase predPhase;
    private final Parser.Table.State state;
    private  boolean fromEmptyReduction;
    private       FastSet<ResultNode> results = new FastSet<ResultNode>();
    private       FastSet<ResultNode> successors = new FastSet<ResultNode>();
    //private       HashSet<ResultNode> results = new HashSet<ResultNode>();
    //private       HashSet<ResultNode> successors = new HashSet<ResultNode>();

    public final void invoke(Pos r, ResultNode only, Object o) {
        boolean emptyProductions = only==null;
        if (emptyProductions != (r.numPops()==0)) return;
        if (r.numPops()!=0)  reduce(r, r.numPops()-1, phase(), only);
        else {
            Input.Region region = phase().getLocation().createRegion(phase().getLocation());
            phase().newNodeFromReduction(r.rewrite(region), r, this);
        }
    }

    private void reduce(Pos r, int pos, GSS.Phase target, ResultNode only) {
        for(ResultNode res : results)
            if (only == null || res == only)
                for(StateNode pred : res.getPreds())
                    reduce2(r, pos, target, pred, res.getForest());
    }

    void reduce2(Pos r, int pos, GSS.Phase target, StateNode pred, Forest f) {
        Forest[] holder = r.holder;
        Forest old = pos >= holder.length ? null : holder[pos];
        if (pos < holder.length) holder[pos] = f;
        if (pos>0)  pred.reduce(r, pos-1, target, null);
        else {
            Input.Region region = pred.phase().getLocation().createRegion(target.getLocation());
            new Reduction(pred, r, r.rewrite(region), target);
        }
        if (pos < holder.length) holder[pos] = old;
    }

    StateNode(GSS.Phase phase, Forest f, Pos reduction, StateNode pred, State state, boolean fromEmptyReduction) {
        this(phase, new ResultNode(f, reduction, pred), state, fromEmptyReduction);
    }
    StateNode(GSS.Phase phase, ResultNode pred, State state, boolean fromEmptyReduction) {
        this.phase = phase;
        this.state = state;
        this.fromEmptyReduction = fromEmptyReduction;
        if (phase.hash.get(state, pred.phase()) != null) throw new Error("severe problem!");
        this.predPhase = pred.phase();
        phase.hash.put(state, pred.phase(), this);

        results.add(pred);
        pred.addSucc(this);
        if (!fromEmptyReduction)
            state.invokeReductions(phase().getToken(), this, pred);

        state.invokeEpsilonReductions(phase().token, this);
    }

    // Add/Remove Successors/Results //////////////////////////////////////////////////////////////////////////////

    public void removeSucc(ResultNode succ)   {
        successors.remove(succ);
        check();
    }
    public void removeResult(ResultNode result) {
        results.remove(result);
        check();
    }
    public void addSucc(ResultNode succ)      {
        successors.add(succ);
    }
    public void addResult(Forest f, Pos reduction, StateNode pred) {
        for(ResultNode r : results)
            if (r.predecessorsContains(pred)) {
                r.merge(f);
                return;
            }
        ResultNode result = new ResultNode(f, reduction, pred);
        results.add(result);
        result.addSucc(this);
        if (!this.fromEmptyReduction) state.invokeReductions(phase().getToken(), this, result);
    }

    // GraphViz //////////////////////////////////////////////////////////////////////////////

    public GraphViz.StateNode toGraphViz(GraphViz gv) {
        if (results.size()==0) return null;
        if (gv.hasNode(this)) return gv.createNode(this);
        GraphViz.StateNode n = gv.createNode(this);
        n.label = "state["+state.toInt()+"]";
        n.shape = "rectangle";
        boolean haspreds = false;
        for(ResultNode r : results) n.edge(r, "");
        n.color = state.doomed ? "red" : "green";
        ((GraphViz.Group)phase().toGraphViz(gv)).add(n);
        return n;
    }
    public boolean isTransparent() { return false; }
    public boolean isHidden() { return false; }

}
